/*
 * Tabs are 4 symbols wide due to stupid Windows function names...
 * ---------------------------------------------------------------
 *
 * Copyright (c) 2010 KSI Labs, LLC. All rights reserved.
 * Written by Sergey Kubushyn <ksi@koi8.net>, <ksi@ksilabs.com>.
 *
 * File: power.c
 *
 * This is Power Management routines for WDM driver for KSI Labs, LLC
 * CENTIPEDE PCI Board. As a matter of fact all this is just a boat anchor
 * because there is absolutely no need for all this most of the times--no
 * one is going to insert/remove a PCI board on a working machine, no one
 * is going to use such a board on a battery-powered laptop, no one is
 * going to powerdown parts of the computer controlling a big dangerous
 * machine--but that is here to comply with braindead Windows concept of
 * "Universal Server of Everything." Most of the code is boilerplate take
 * almost verbatim from DDK examples and don't serve any reasonable purpose.
 *
 * Please read documentation or go to http://www.ksilabs.com for further info.
 */

/*
 * 						LICENSE and Terms of Use.
 *
 * First of all, this is NOT GPL or Free Software. It is open so you have
 * all the source code for the Engine. There is also VHDL source code for
 * CENTIPEDE board CPLD that is available from KSI Labs, LLC website,
 * http://www.ksilabs.com.
 *
 * As for this particular code, Windows Mach3 Engine, you are free to use it,
 * make any modifications to customize it for your own needs, fix inevitable
 * bugs, port it to other CNC software and do anything else as long as your
 * derivative work that is based on this code or uses parts or pieces thereof
 * is made to be used with any KSI Labs, LLC products.
 *
 * We do NOT grant you a permission to use any parts of this software in any
 * products that are designed for products not manufactured or distributed
 * by KSI Labs, LLC. We might give such a permission on case-by-case basis
 * but it is not legal to use this software or bits and pieces thereof with
 * products not coming from KSI Labs, LLC without our explicit written
 * permission.
 *
 * The original copyright notices must be included in all derivative works.
 * You do NOT have to ask our permission to make any modifications to this
 * code and you are not under obligation to give us any of your work back
 * as long as you are using it for your own needs.
 *
 * However we do NOT give our permission to distribute your derivative work
 * unless all original copyrights are retained, all the conditions of this
 * license are met (use only with KSI Labs, LLC products) and full source
 * code for such derivative work is available for free to everybody without
 * any further conditions including but not limited to signing any kinds of
 * NDA. We explicitely deny permission to distribute any derivative work
 * based on this code with closed source binary-only fragments or include
 * bits or pieces of this code in such products.
 *
 * CNC software like Mach3 and likes is excluded from these requirements as
 * long as it does not include parts or pieces of this software into the
 * close source proprietary application proper. It can use shared memory
 * buffer, IOCTL calls, etc. to communicate with this Engine or the one based
 * on it and keep all the software making those calls or working on that
 * buffer close source and proprietary as long as those calls/structures are
 * documented and that documentation is freely available to everybody without
 * any further conditions including but no limited to signing any kind of NDA.
 */

#include "precomp.h"
#include "power.h"

#ifdef ALLOC_PRAGMA
#pragma alloc_text(PAGE_PWR, DispatchPower)
#pragma alloc_text(PAGE_PWR, DispatchPowerDefault)
#pragma alloc_text(PAGE_PWR, DispatchSetPowerState)
#pragma alloc_text(PAGE_PWR, DispatchQueryPowerState)
#pragma alloc_text(PAGE_PWR, DispatchSystemPowerIrp)
#pragma alloc_text(PAGE_PWR, DispatchDeviceQueryPower)
#pragma alloc_text(PAGE_PWR, DispatchDeviceSetPower)
#pragma alloc_text(PAGE_PWR, PowerBeginQueuingIrps)
#pragma alloc_text(PAGE_PWR, CallbackHandleDeviceQueryPower)
#pragma alloc_text(PAGE_PWR, CallbackHandleDeviceSetPower)
#endif // ALLOC_PRAGMA

//
// The code for this file is split into two sections, boilerplate code and
// device programming code.
//
// The Boilerplate code comes first. The device programming code comes last
// in the file.
//

/////////////////////////////////////////////////////////////////////////
// There might be bugs in this file but you should blame Microsoft for //
// them :) All this file as well as corresponding header, power.h is   //
// ABSOLUTELY useless ballast doing nothing except occupying precious  //
// kernel memory. Other than fixing a possible bug there is ABSOLUTELY //
// no need to even look at this. It serves NO PURPOSE.                 //
/////////////////////////////////////////////////////////////////////////

// ----------------------
// Begin Boilerplate Code
// ----------------------


// *******************************
// * Power Dispatch main routine *
// *******************************
NTSTATUS DispatchPower(PDEVICE_OBJECT DeviceObject, PIRP Irp)
{
	PIO_STACK_LOCATION  stack;
	PFDO_DATA           FdoData;
	NTSTATUS            status;

	stack   = IoGetCurrentIrpStackLocation(Irp);
	FdoData = (PFDO_DATA)DeviceObject->DeviceExtension;

	ASSERT(stack->MajorFunction == IRP_MJ_POWER);

	DebugPrint(INFO, DBG_POWER, "FDO %s %s IRP:0x%p %s %s\n",
				DbgPowerMinorFunctionString(stack->MinorFunction),
				stack->Parameters.Power.Type == SystemPowerState ?"SIRP":"DIRP",
				Irp,
				DbgSystemPowerString(FdoData->SystemPowerState),
				DbgDevicePowerString(FdoData->DevicePowerState));

	IoIncrement (FdoData);

	if (FdoData->DevicePnPState == Deleted) {

		PoStartNextPowerIrp(Irp);
		Irp->IoStatus.Status = STATUS_NO_SUCH_DEVICE;
		IoCompleteRequest(Irp, IO_NO_INCREMENT);
		IoDecrement(FdoData);

		return STATUS_NO_SUCH_DEVICE;
	}

	if (FdoData->DevicePnPState == NotStarted) {

		PoStartNextPowerIrp(Irp);
		IoSkipCurrentIrpStackLocation(Irp);
		status = PoCallDriver(FdoData->NextLowerDriver, Irp);
		IoDecrement(FdoData);

		return status;
	}

	switch (stack->MinorFunction) {

		case IRP_MN_SET_POWER:

			status = DispatchSetPowerState(DeviceObject, Irp);

			break;

		case IRP_MN_QUERY_POWER:

			status = DispatchQueryPowerState(DeviceObject, Irp);

			break;

		case IRP_MN_WAIT_WAKE:
			//
			// The minor power IRP code IRP_MN_WAIT_WAKE provides
			// for waking a device or waking the system. Drivers
			// of devices that can wake themselves or the system
			// send IRP_MN_WAIT_WAKE. The system sends IRP_MN_WAIT_WAKE
			// only to devices that always wake the system, such as
			// the power-on switch.
			//
			// KSI: We do NOT support this

		case IRP_MN_POWER_SEQUENCE:
			//
			// A driver sends this IRP as an optimization to determine
			// whether its device actually entered a specific power state.
			// This IRP is optional. Power Manager cannot send this IRP.
			//
			// KSI: We do NOT support this

		default:
			//
			// Pass it down
			//
			status = DispatchPowerDefault(DeviceObject, Irp);
			IoDecrement(FdoData);
			break;
	}

	return status;
}


// ***********************************************
// * Sends unsupported Power IRPs down the stack *
// ***********************************************
NTSTATUS DispatchPowerDefault(PDEVICE_OBJECT DeviceObject, PIRP Irp)
{
	NTSTATUS	status;
	PFDO_DATA	FdoData;

	PoStartNextPowerIrp(Irp);

	IoSkipCurrentIrpStackLocation(Irp);

	FdoData = (PFDO_DATA)DeviceObject->DeviceExtension;

	status = PoCallDriver(FdoData->NextLowerDriver, Irp);

	return status;
}


// **************************************
// * Dispatch routine for SetPower IRPs *
// **************************************
NTSTATUS DispatchSetPowerState(PDEVICE_OBJECT DeviceObject, PIRP Irp)
{
	PIO_STACK_LOCATION	stack = IoGetCurrentIrpStackLocation(Irp);

	DebugPrint(TRACE, DBG_POWER, "--> DispatchSetPowerState\n");

	return (stack->Parameters.Power.Type == SystemPowerState) ?
						DispatchSystemPowerIrp(DeviceObject, Irp) :
							DispatchDeviceSetPower(DeviceObject, Irp);
}


// ****************************************
// * Dispatch routine for QueryPower IRPs *
// ****************************************
NTSTATUS DispatchQueryPowerState(PDEVICE_OBJECT DeviceObject, PIRP Irp)
{
	PIO_STACK_LOCATION	stack = IoGetCurrentIrpStackLocation(Irp);

	DebugPrint(TRACE, DBG_POWER, "--> DispatchQueryPowerState\n");

	return (stack->Parameters.Power.Type == SystemPowerState) ?
						DispatchSystemPowerIrp(DeviceObject, Irp) :
							DispatchDeviceQueryPower(DeviceObject, Irp);
}


// **********************************************************************
// * Dispatch routine for SYSTEM SetPower AND QueryPower IRPs (S-state) *
// **********************************************************************
NTSTATUS DispatchSystemPowerIrp(PDEVICE_OBJECT DeviceObject, PIRP Irp)
{
	PIO_STACK_LOCATION  stack = IoGetCurrentIrpStackLocation(Irp);
	PFDO_DATA           FdoData = (PFDO_DATA) DeviceObject->DeviceExtension;

	DebugPrint(TRACE, DBG_POWER, "--> DispatchSystemPowerIrp\n");

	if (stack->MinorFunction == IRP_MN_SET_POWER) {
		FdoData->SystemPowerState = stack->Parameters.Power.State.SystemState;
		DebugPrint(INFO, DBG_POWER, "\tSetting the system state to %s\n",
					DbgSystemPowerString(FdoData->SystemPowerState));
	}

	IoMarkIrpPending(Irp);

	IoCopyCurrentIrpStackLocationToNext(Irp);

	IoSetCompletionRoutine(Irp, (PIO_COMPLETION_ROUTINE)CompletionSystemPowerUp,
							NULL, TRUE, TRUE, TRUE);

	PoCallDriver(FdoData->NextLowerDriver, Irp);

	return STATUS_PENDING;
}


// *************************************************************
// * Completion routine for PowerUp S-IRP hooked up by the     *
// * DispatchSystemPowerIrp() routine above. It completes      *
// * the S-IRP and initiates the corresponding D-IRP by        *
// * calling yet another routine to actually queue that D-IRP. *
// *************************************************************
NTSTATUS CompletionSystemPowerUp(PDEVICE_OBJECT Fdo, PIRP Irp, PVOID NotUsed)
{
	PFDO_DATA	FdoData = (PFDO_DATA)Fdo->DeviceExtension;
	NTSTATUS	status = Irp->IoStatus.Status;

	DebugPrint(TRACE, DBG_POWER, "--> CompletionSystemPowerUp\n");

	if (!NT_SUCCESS(status)) {

		PoStartNextPowerIrp(Irp);
		IoDecrement(FdoData);

		return STATUS_CONTINUE_COMPLETION;
	}

	QueueCorrespondingDeviceIrp(Irp, Fdo);

	return STATUS_MORE_PROCESSING_REQUIRED;
}


// *****************************************************************
// * This queues the corresponding D-IRP for an S-IRP. Called from *
// * CompletionSystemPowerUp() routine above. But that is not the  *
// * end of a story yet -- it sends that D-IRP down the stack and  *
// * sets yet another completion routine to complete that D-IRP,   *
// * CompletionOnFinalizedDeviceIrp() below...                     *
// *****************************************************************
VOID QueueCorrespondingDeviceIrp(PIRP SIrp, PDEVICE_OBJECT DeviceObject)
{
	NTSTATUS            status;
	POWER_STATE         state;
	PFDO_DATA           FdoData = (PFDO_DATA)DeviceObject->DeviceExtension;
	PIO_STACK_LOCATION  stack = IoGetCurrentIrpStackLocation(SIrp);
	UCHAR               minor = stack->MinorFunction;
	BOOLEAN             sIrpCompleted = FALSE;

	DebugPrint(TRACE, DBG_POWER, "--> QueueCorrespondingDeviceIrp\n");

	//
	// Get the device state to request.
	//
	status = GetPowerPoliciesDeviceState(SIrp, DeviceObject, &state.DeviceState);

	if (NT_SUCCESS(status)) {

		ASSERT(!FdoData->PendingSIrp);
		FdoData->PendingSIrp = SIrp;

		if (stack->MinorFunction == IRP_MN_SET_POWER &&
			stack->Parameters.Power.State.SystemState == PowerSystemWorking) {

			DebugPrint(TRACE, DBG_POWER, "\t ****Completing S0 IRP\n");

			SIrp->IoStatus.Status = STATUS_SUCCESS;

			ASSERT(state.DeviceState == PowerDeviceD0);

			IoIncrement(FdoData);
			FdoData->PendingSIrp = (PVOID)MAGIC_NUMBER;
			PoStartNextPowerIrp(SIrp);
			IoCompleteRequest(SIrp, IO_NO_INCREMENT);
			IoDecrement(FdoData);
			sIrpCompleted = TRUE;
		}

		status = PoRequestPowerIrp(FdoData->UnderlyingPDO, minor, state,
							CompletionOnFinalizedDeviceIrp, FdoData, NULL);
	}

	if (!NT_SUCCESS(status)) {

		FdoData->PendingSIrp = NULL;

		if (sIrpCompleted) {

			IoDecrement(FdoData);

		} else {

			PoStartNextPowerIrp(SIrp);
			SIrp->IoStatus.Status = status;
			IoCompleteRequest(SIrp, IO_NO_INCREMENT);
			IoDecrement(FdoData);

		}
	}
}


// *******************************************
// * Completion routine for finalizing D-IRP *
// * that came back from the lower layer.    *
// *******************************************
VOID CompletionOnFinalizedDeviceIrp(PDEVICE_OBJECT DeviceObject, UCHAR MinorFunction,
		POWER_STATE PowerState, PVOID PowerContext, PIO_STATUS_BLOCK IoStatus)
{
	PFDO_DATA	FdoData = (PFDO_DATA)PowerContext;
	PIRP		sIrp = FdoData->PendingSIrp;

	DebugPrint(TRACE, DBG_POWER, "--> CompletionOnFinalizedDeviceIrp\n");

	if (sIrp == (PVOID)MAGIC_NUMBER) {

		ASSERT(PowerState.DeviceState == PowerDeviceD0);
		sIrp = FdoData->PendingSIrp = NULL;
		IoDecrement(FdoData);
		return;
	}

	if (sIrp) {

		sIrp->IoStatus.Status = IoStatus->Status;
		PoStartNextPowerIrp(sIrp);
		IoCompleteRequest(sIrp, IO_NO_INCREMENT);
		FdoData->PendingSIrp = NULL;
		IoDecrement(FdoData);
	}
}


// ****************************************
// * Handles IRP_MN_QUERY_POWER for D-IRP *
// ****************************************
NTSTATUS DispatchDeviceQueryPower(PDEVICE_OBJECT DeviceObject, PIRP Irp)
{
	PIO_STACK_LOCATION  stack = IoGetCurrentIrpStackLocation(Irp);
	DEVICE_POWER_STATE  deviceState = stack->Parameters.Power.State.DeviceState;
	NTSTATUS            status;

	DebugPrint(TRACE, DBG_POWER, "--> DispatchDeviceQueryPower\n");

	if (deviceState == PowerDeviceD0) {

		status = STATUS_SUCCESS;

	} else {

		status = QueuePassiveLevelCallback(
					(PFDO_DATA)DeviceObject->DeviceExtension,
					CallbackHandleDeviceQueryPower,
					(PVOID)Irp, // Pointer to D-IRP
					(PVOID)IRP_NEEDS_FORWARDING // Directio
					);

		if (NT_SUCCESS(status))
			return STATUS_PENDING;
	}

	return FinalizeDevicePowerIrp(DeviceObject, Irp, IRP_NEEDS_FORWARDING, status);
}


// **************************************
// * Handles IRP_MN_SET_POWER for D-IRP *
// **************************************
NTSTATUS DispatchDeviceSetPower(PDEVICE_OBJECT DeviceObject, PIRP Irp)
{
	PIO_STACK_LOCATION  stack = IoGetCurrentIrpStackLocation(Irp);
	POWER_STATE         state = stack->Parameters.Power.State;
	NTSTATUS            status;
	PFDO_DATA           FdoData = (PFDO_DATA)DeviceObject->DeviceExtension;

	DebugPrint(TRACE, DBG_POWER, "--> DispatchDeviceSetPower\n");

	if (state.DeviceState < FdoData->DevicePowerState) { // adding power

		IoMarkIrpPending(Irp);

		IoCopyCurrentIrpStackLocationToNext(Irp);

		IoSetCompletionRoutine(Irp,
			(PIO_COMPLETION_ROUTINE)CompletionDevicePowerUp,
										NULL, TRUE, TRUE, TRUE);

		PoCallDriver(FdoData->NextLowerDriver, Irp);

		return STATUS_PENDING;

	} else {

		status = QueuePassiveLevelCallback(FdoData,
								CallbackHandleDeviceSetPower,
								(PVOID)Irp, (PVOID)IRP_NEEDS_FORWARDING);

		if (NT_SUCCESS(status))
			return STATUS_PENDING;

		ASSERT(!NT_SUCCESS(status));

		return FinalizeDevicePowerIrp(DeviceObject, Irp, IRP_NEEDS_FORWARDING, status);
	}
}


// **********************************************
// * The completion routine for Power Up D-IRP. *
// **********************************************
NTSTATUS CompletionDevicePowerUp(PDEVICE_OBJECT DeviceObject, PIRP Irp, PVOID NotUsed)
{
	PFDO_DATA           FdoData = (PFDO_DATA)DeviceObject->DeviceExtension;
	NTSTATUS            status = Irp->IoStatus.Status;
	PIO_STACK_LOCATION  stack = IoGetCurrentIrpStackLocation(Irp);

	DebugPrint(TRACE, DBG_POWER, "--> CompletionDevicePowerUp\n");

	if (!NT_SUCCESS(status)) {

		PoStartNextPowerIrp(Irp);
		IoDecrement(FdoData);
		return STATUS_CONTINUE_COMPLETION;
	}

	ASSERT(stack->MajorFunction == IRP_MJ_POWER);
	ASSERT(stack->MinorFunction == IRP_MN_SET_POWER);

	status = QueuePassiveLevelCallback(FdoData,
							CallbackHandleDeviceSetPower,
							(PVOID)Irp, (PVOID)IRP_ALREADY_FORWARDED);

	if (!NT_SUCCESS(status)) {

		FinalizeDevicePowerIrp(DeviceObject, Irp, IRP_ALREADY_FORWARDED, status);
	}

	return STATUS_MORE_PROCESSING_REQUIRED;
}


// *********************************************
// * This is the final step in D-IRP handling. *
// *********************************************
NTSTATUS FinalizeDevicePowerIrp(PDEVICE_OBJECT DeviceObject, PIRP Irp,
								IRP_DIRECTION Direction, NTSTATUS Result)
{
	NTSTATUS	status;
	PFDO_DATA	FdoData = (PFDO_DATA)DeviceObject->DeviceExtension;

	DebugPrint(TRACE, DBG_POWER, "--> FinalizeDevicePowerIrp\n");

	if (Direction == IRP_ALREADY_FORWARDED || (!NT_SUCCESS(Result))) {

		PoStartNextPowerIrp(Irp);
		Irp->IoStatus.Status = Result;
		IoCompleteRequest(Irp, IO_NO_INCREMENT);
		IoDecrement(FdoData);

		return Result;
	}

	Irp->IoStatus.Status = Result;
	status = DispatchPowerDefault(DeviceObject, Irp);
	IoDecrement(FdoData);

	return status;
}


// ****************************************************************
// * Handles the actual query-power state changes for the device. *
// ****************************************************************
VOID CallbackHandleDeviceQueryPower(PDEVICE_OBJECT DeviceObject,
										PWORK_ITEM_CONTEXT Context)
{
	DEVICE_POWER_STATE	deviceState;
	PIO_STACK_LOCATION	stack;
	NTSTATUS			status;
	PIRP				irp;
	IRP_DIRECTION		direction;
	PFDO_DATA			FdoData = (PFDO_DATA)DeviceObject->DeviceExtension;

	PAGED_CODE();

	irp = Context->Argument1;
	direction = (IRP_DIRECTION)Context->Argument2;
	stack = IoGetCurrentIrpStackLocation(irp);
	deviceState = stack->Parameters.Power.State.DeviceState;

	ASSERT(PowerDeviceD0 != deviceState);

	status = PowerBeginQueuingIrps(DeviceObject, FdoData->PendingSIrp? 2 : 1, TRUE);

	FinalizeDevicePowerIrp(DeviceObject, irp, direction, status);
	IoFreeWorkItem((PIO_WORKITEM)Context->WorkItem);
	ExFreePool((PVOID)Context);
}


// ****************************************************************
// * Start queueing IRPs and finish currently pending ones.       *
// * NB: It checks if Query is set and checks if it is allowed    *
// *     to suspend the device. If it is not it restarts the      *
// *     queue and returns FALSE. CENTIPEDE can NOT be suspended. *
// ****************************************************************
NTSTATUS PowerBeginQueuingIrps(PDEVICE_OBJECT DeviceObject, ULONG IrpIoCharges,
													BOOLEAN Query)
{
	NTSTATUS	status;
	PFDO_DATA	FdoData;

	PAGED_CODE();

	FdoData = (PFDO_DATA)DeviceObject->DeviceExtension;

	if (Query) {

		status = CanSuspendDevice(DeviceObject);

		if (!NT_SUCCESS(status))
			return status;

	}

	status = STATUS_SUCCESS;

	FdoData->QueueState = HoldRequests;

	ReleaseAndWait(FdoData, IrpIoCharges, DEV_STOP);

	return status;
}


// --------------------------
// Begin device-specific code
// --------------------------


// ****************************************************
// * Performs the actual power changes to the device. *
// ****************************************************
VOID CallbackHandleDeviceSetPower(PDEVICE_OBJECT DeviceObject,
										PWORK_ITEM_CONTEXT Context)
{
	PFDO_DATA			FdoData = (PFDO_DATA)DeviceObject->DeviceExtension;
	POWER_ACTION		newDeviceAction;
	DEVICE_POWER_STATE	newDeviceState, oldDeviceState;
	POWER_STATE			newState;
	NTSTATUS			status = STATUS_SUCCESS;
	PIRP				Irp = Context->Argument1;
	IRP_DIRECTION		Direction = (IRP_DIRECTION)Context->Argument2;
	PIO_STACK_LOCATION	stack = IoGetCurrentIrpStackLocation(Irp);

	DebugPrint(TRACE, DBG_POWER, "--> CallbackHandleDeviceSetPower\n");

	PAGED_CODE();

	newState =  stack->Parameters.Power.State;
	newDeviceState = newState.DeviceState;
	oldDeviceState = FdoData->DevicePowerState;

	if (newDeviceState == PowerDeviceD0 &&
		FdoData->SystemPowerState != PowerSystemWorking) {

		FinalizeDevicePowerIrp(DeviceObject, Irp, Direction, STATUS_UNSUCCESSFUL);

		goto Exit;
	}

	DebugPrint(INFO, DBG_POWER, "\tSetting the device state from %s to %s\n",
							DbgDevicePowerString(oldDeviceState),
							DbgDevicePowerString(newDeviceState));

	FdoData->DevicePowerState = newDeviceState;

	if (newDeviceState == oldDeviceState) {

		DebugPrint(INFO, DBG_POWER, "\tNo Change in Power state\n");

		if (newDeviceState == PowerDeviceD0) {
			FdoData->QueueState = AllowRequests;
			ProcessQueuedRequests(FdoData);
		}

		FinalizeDevicePowerIrp(DeviceObject, Irp, Direction, STATUS_SUCCESS);

		goto Exit;
	}

	if (oldDeviceState == PowerDeviceD0) {

		status = PowerBeginQueuingIrps(DeviceObject, FdoData->PendingSIrp? 2 : 1, FALSE);

		ASSERT(NT_SUCCESS(status));
	}

	status = HWSetPower(FdoData, newDeviceState);	// That is a dummy
	ASSERTMSG("HWSetPower failed\n", status == STATUS_SUCCESS);

	PoSetPowerState(DeviceObject, DevicePowerState, newState);

	newDeviceAction = stack->Parameters.Power.ShutdownType;

	if (newDeviceState > oldDeviceState) {

		if (newDeviceAction >= PowerActionShutdown)
			HWShutdown(FdoData);
	}

	if (newDeviceState == PowerDeviceD0) {
		FdoData->QueueState = AllowRequests;
		ProcessQueuedRequests(FdoData);
	}

	FinalizeDevicePowerIrp(DeviceObject, Irp, Direction, status);

Exit:
	IoFreeWorkItem((PIO_WORKITEM)Context->WorkItem);
	ExFreePool((PVOID)Context);
}


// ***********************************************************
// * Retrieves the appropriate D state for a given S-IRP.    *
// * This is very simple implementation for CENTIPEDE board. *
// ***********************************************************
NTSTATUS GetPowerPoliciesDeviceState(PIRP SIrp, PDEVICE_OBJECT DeviceObject,
										DEVICE_POWER_STATE *DevicePowerState)
{
	PIO_STACK_LOCATION	stack = IoGetCurrentIrpStackLocation(SIrp);
	SYSTEM_POWER_STATE	systemState = stack->Parameters.Power.State.SystemState;

	if (systemState == PowerSystemWorking)
		*DevicePowerState = PowerDeviceD0;
	else
		*DevicePowerState = PowerDeviceD3;

	return STATUS_SUCCESS;
}


// *********************************************************
// * Determines whether a device powerdown request should  *
// * succeeded or failed. Always fail for CENTIPEDE board. *
// *********************************************************
NTSTATUS CanSuspendDevice(PDEVICE_OBJECT DeviceObject)
{
	return STATUS_UNSUCCESSFUL;
}
